@@ -64,12 +64,14 @@
 	    [ print_html/2, print_html/1, page/4, html/3,
 	      html_print_length/2
 	    ]).
+:- if(exists_source(http_exception)).
 :- autoload(http_exception,[map_exception_to_http_status/4]).
+:- endif.
 :- autoload(mimepack,[mime_pack/3]).
 :- autoload(mimetype,[file_mime_type/2]).
 :- autoload(library(apply),[maplist/2]).
 :- autoload(library(base64),[base64/2]).
-:- autoload(library(debug),[debug/3,debugging/1]).
+:- use_module(library(debug),[debug/3,debugging/1]).
 :- autoload(library(error),[syntax_error/1,domain_error/2]).
 :- autoload(library(lists),[append/3,member/2,select/3,delete/3]).
 :- autoload(library(memfile),
@@ -248,12 +250,18 @@ http_reply(Data, Out, HdrExtra, _Context, Request, Code) :-
         throw(error(http_write_short(Data, Sent), _))
     ;   E = error(timeout_error(write, _), _)
     ->  throw(E)
-    ;   map_exception_to_http_status(E, Status, NewHdr, NewContext),
-        http_status_reply(Status, Out, NewHdr, NewContext, Request, Code)
+    ;   map_exception_to_http_status(E, Status, NewHdr, NewContext)
+    ->  http_status_reply(Status, Out, NewHdr, NewContext, Request, Code)
+    ;   throw(E)
     ).
 http_reply(Status, Out, HdrExtra, Context, Request, Code) :-
     http_status_reply(Status, Out, HdrExtra, Context, Request, Code).
 
+:- if(\+current_predicate(map_exception_to_http_status/4)).
+map_exception_to_http_status(_E, _Status, _NewHdr, _NewContext) :-
+    fail.
+:- endif.
+
 :- meta_predicate
     if_no_head(0, +).
 
@@ -271,7 +279,7 @@ http_reply_data(Data, Out, HdrExtra, Method, Code) :-
 http_reply_data_(html(HTML), Out, HdrExtra, Method, Code) :-
     !,
     phrase(reply_header(html(HTML), HdrExtra, Code), Header),
-    format(Out, '~s', [Header]),
+    send_reply_header(Out, Header),
     if_no_head(print_html(Out, HTML), Method).
 http_reply_data_(file(Type, File), Out, HdrExtra, Method, Code) :-
     !,
@@ -292,7 +300,7 @@ http_reply_data_(tmp_file(Type, File), Out, HdrExtra, Method, Code) :-
 http_reply_data_(bytes(Type, Bytes), Out, HdrExtra, Method, Code) :-
     !,
     phrase(reply_header(bytes(Type, Bytes), HdrExtra, Code), Header),
-    format(Out, '~s', [Header]),
+    send_reply_header(Out, Header),
     if_no_head(format(Out, '~s', [Bytes]), Method).
 http_reply_data_(stream(In, Len), Out, HdrExtra, Method, Code) :-
     !,
@@ -314,7 +322,7 @@ if_no_head(Goal, _) :-
 
 reply_file(Out, _File, Header, head) :-
     !,
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 reply_file(Out, File, Header, _) :-
     setup_call_cleanup(
         open(File, read, In, [type(binary)]),
@@ -323,7 +331,7 @@ reply_file(Out, File, Header, _) :-
 
 reply_file_range(Out, _File, Header, _Range, head) :-
     !,
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 reply_file_range(Out, File, Header, bytes(From, To), _) :-
     setup_call_cleanup(
         open(File, read, In, [type(binary)]),
@@ -332,7 +340,7 @@ reply_file_range(Out, File, Header, bytes(From, To), _) :-
 
 copy_stream(Out, _, Header, head, _, _) :-
     !,
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 copy_stream(Out, In, Header, _, From, To) :-
     copy_stream(Out, In, Header, From, To).
 
@@ -342,7 +350,7 @@ copy_stream(Out, In, Header, From, To) :-
     ;   seek(In, From, bof, _)
     ),
     peek_byte(In, _),
-    format(Out, '~s', [Header]),
+    send_reply_header(Out, Header),
     (   To == end
     ->  copy_stream_data(In, Out)
     ;   Len is To - From,
@@ -416,7 +424,7 @@ status_reply_flush(Status, Out, Options) :-
     !,
     flush_output(Out).
 
-%!  status_reply(+Status, +Out, +Options:Dict)
+%!  status_reply(+Status, +Out, +Options:dict)
 %
 %   Formulate a non-200 reply and send it to the stream Out.  Options
 %   is a dict containing:
@@ -431,7 +439,7 @@ status_reply_flush(Status, Out, Options) :-
 status_reply(no_content, Out, Options) :-
     !,
     phrase(reply_header(status(no_content), Options), Header),
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 status_reply(switching_protocols(_Goal,SwitchOptions), Out, Options) :-
     !,
     (   option(headers(Extra1), SwitchOptions)
@@ -441,7 +449,7 @@ status_reply(switching_protocols(_Goal,SwitchOptions), Out, Options) :-
     http_join_headers(Options.header, Extra1, HdrExtra),
     phrase(reply_header(status(switching_protocols),
                         Options.put(header,HdrExtra)), Header),
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 status_reply(authorise(basic, ''), Out, Options) :-
     !,
     status_reply(authorise(basic), Out, Options).
@@ -451,7 +459,7 @@ status_reply(authorise(basic, Realm), Out, Options) :-
 status_reply(not_modified, Out, Options) :-
     !,
     phrase(reply_header(status(not_modified), Options), Header),
-    format(Out, '~s', [Header]).
+    send_reply_header(Out, Header).
 % aliases (compatibility)
 status_reply(busy, Out, Options) :-
     status_reply(service_unavailable(busy), Out, Options).
@@ -468,7 +476,7 @@ status_reply(Status, Out, Options) :-
     append(List, [Body], ExList),
     ExStatus =.. ExList,
     phrase(reply_header(ExStatus, Options), Header),
-    format(Out, '~s', [Header]),
+    send_reply_header(Out, Header),
     reply_status_body(Out, Body, Options).
 
 %!  status_has_content(+StatusTerm, -HTTPCode)
@@ -479,6 +487,7 @@ status_reply(Status, Out, Options) :-
 status_has_content(created(_Location)).
 status_has_content(moved(_To)).
 status_has_content(moved_temporary(_To)).
+status_has_content(gone(_URL)).
 status_has_content(see_other(_To)).
 status_has_content(bad_request(_ErrorTerm)).
 status_has_content(authorise(_Method)).
@@ -589,6 +598,16 @@ status_page_hook(moved_temporary(To), html_tokens(HTML), _Options) :-
                   \address
                 ]),
            HTML).
+status_page_hook(gone(URL), html_tokens(HTML), _Options) :-
+    phrase(page([ title('410 Resource Gone')
+                ],
+                [ h1('Resource Gone'),
+                  p(['The document has been removed ',
+                     a(href(URL), ' from here')
+                    ]),
+                  \address
+                ]),
+           HTML).
 status_page_hook(see_other(To), html_tokens(HTML), _Options) :-
     phrase(page([ title('303 See Other')
                  ],
@@ -696,6 +715,20 @@ html_message_lines([nl|T]) -->
     html_message_lines(T).
 html_message_lines([flush]) -->
     [].
+html_message_lines([ansi(_Style,Fmt,Args)|T]) -->
+    !,
+    { format(string(S), Fmt, Args)
+    },
+    html([S]),
+    html_message_lines(T).
+html_message_lines([url(Pos)|T]) -->
+    !,
+    msg_url(Pos),
+    html_message_lines(T).
+html_message_lines([url(URL, Label)|T]) -->
+    !,
+    html(a(href(URL), Label)),
+    html_message_lines(T).
 html_message_lines([Fmt-Args|T]) -->
     !,
     { format(string(S), Fmt, Args)
@@ -709,6 +742,15 @@ html_message_lines([Fmt|T]) -->
     html([S]),
     html_message_lines(T).
 
+msg_url(File:Line:Pos) -->
+    !,
+    html([File, :, Line, :, Pos]).
+msg_url(File:Line) -->
+    !,
+    html([File, :, Line]).
+msg_url(File) -->
+    html([File]).
+
 %!  http_join_headers(+Default, +Header, -Out)
 %
 %   Append headers from Default to Header if they are not
@@ -745,9 +787,7 @@ http_update_encoding(Header0, utf8, [content_type(Type)|Header]) :-
     atom_concat(B, '; charset=UTF-8', Type).
 http_update_encoding(Header, Encoding, Header) :-
     memberchk(content_type(Type), Header),
-    (   (   sub_atom(Type, _, _, _, 'UTF-8')
-        ;   sub_atom(Type, _, _, _, 'utf-8')
-        )
+    (   sub_atom_icasechk(Type, _, 'utf-8')
     ->  Encoding = utf8
     ;   http:mime_type_encoding(Type, Encoding)
     ->  true
@@ -760,13 +800,15 @@ http_update_encoding(Header, octet, Header).
 %   Encoding is the (default) character encoding for MimeType. Hooked by
 %   http:mime_type_encoding/2.
 
-mime_type_encoding('application/json',         utf8).
-mime_type_encoding('application/jsonrequest',  utf8).
-mime_type_encoding('application/x-prolog',     utf8).
-mime_type_encoding('application/n-quads',      utf8).
-mime_type_encoding('application/n-triples',    utf8).
-mime_type_encoding('application/sparql-query', utf8).
-mime_type_encoding('application/trig',         utf8).
+mime_type_encoding('application/json',                utf8).
+mime_type_encoding('application/jsonrequest',         utf8).
+mime_type_encoding('application/x-prolog',            utf8).
+mime_type_encoding('application/n-quads',             utf8).
+mime_type_encoding('application/n-triples',           utf8).
+mime_type_encoding('application/sparql-query',        utf8).
+mime_type_encoding('application/trig',                utf8).
+mime_type_encoding('application/sparql-results+json', utf8).
+mime_type_encoding('application/sparql-results+xml',  utf8).
 
 %!  http:mime_type_encoding(+MimeType, -Encoding) is semidet.
 %
@@ -908,6 +950,16 @@ content_length_in_encoding(Enc, Stream, Bytes) :-
 %     * html(+Tokens)
 %     Result of html//1 from html_write.pl
 %
+%     * json(+Term)
+%     Posting a JSON query and processing the JSON reply (or any other
+%     reply understood by http_read_data/3) is simple as
+%     =|http_post(URL, json(Term), Reply, [])|=, where Term is a JSON
+%     term as described in json.pl and reply is of the same format if
+%     the server replies with JSON, when using module =|:-
+%     use_module(library(http/http_json))|=. Note that the module is
+%     used in both http server and http client, see
+%     library(http/http_json).
+%
 %     * xml(+Term)
 %     Post the result of xml_write/3 using the Mime-type
 %     =|text/xml|=
@@ -996,7 +1048,7 @@ http_post_data(Data, Out, HdrExtra) :-
 http_post_data(html(HTML), Out, HdrExtra) :-
     !,
     phrase(post_header(html(HTML), HdrExtra), Header),
-    format(Out, '~s', [Header]),
+    send_request_header(Out, Header),
     print_html(Out, HTML).
 http_post_data(xml(XML), Out, HdrExtra) :-
     !,
@@ -1025,7 +1077,7 @@ http_post_data(file(File), Out, HdrExtra) :-
 http_post_data(file(Type, File), Out, HdrExtra) :-
     !,
     phrase(post_header(file(Type, File), HdrExtra), Header),
-    format(Out, '~s', [Header]),
+    send_request_header(Out, Header),
     setup_call_cleanup(
         open(File, read, In, [type(binary)]),
         copy_stream_data(In, Out),
@@ -1033,7 +1085,7 @@ http_post_data(file(Type, File), Out, HdrExtra) :-
 http_post_data(memory_file(Type, Handle), Out, HdrExtra) :-
     !,
     phrase(post_header(memory_file(Type, Handle), HdrExtra), Header),
-    format(Out, '~s', [Header]),
+    send_request_header(Out, Header),
     setup_call_cleanup(
         open_memory_file(Handle, read, In, [encoding(octet)]),
         copy_stream_data(In, Out),
@@ -1044,7 +1096,7 @@ http_post_data(codes(Codes), Out, HdrExtra) :-
 http_post_data(codes(Type, Codes), Out, HdrExtra) :-
     !,
     phrase(post_header(codes(Type, Codes), HdrExtra), Header),
-    format(Out, '~s', [Header]),
+    send_request_header(Out, Header),
     setup_call_cleanup(
         set_stream(Out, encoding(utf8)),
         format(Out, '~s', [Codes]),
@@ -1052,18 +1104,30 @@ http_post_data(codes(Type, Codes), Out, HdrExtra) :-
 http_post_data(bytes(Type, Bytes), Out, HdrExtra) :-
     !,
     phrase(post_header(bytes(Type, Bytes), HdrExtra), Header),
-    format(Out, '~s~s', [Header, Bytes]).
+    send_request_header(Out, Header),
+    format(Out, '~s', [Bytes]).
 http_post_data(atom(Atom), Out, HdrExtra) :-
     !,
     http_post_data(atom(text/plain, Atom), Out, HdrExtra).
 http_post_data(atom(Type, Atom), Out, HdrExtra) :-
     !,
     phrase(post_header(atom(Type, Atom), HdrExtra), Header),
-    format(Out, '~s', [Header]),
+    send_request_header(Out, Header),
     setup_call_cleanup(
         set_stream(Out, encoding(utf8)),
         write(Out, Atom),
         set_stream(Out, encoding(octet))).
+http_post_data(string(String), Out, HdrExtra) :-
+    !,
+    http_post_data(atom(text/plain, String), Out, HdrExtra).
+http_post_data(string(Type, String), Out, HdrExtra) :-
+    !,
+    phrase(post_header(string(Type, String), HdrExtra), Header),
+    send_request_header(Out, Header),
+    setup_call_cleanup(
+        set_stream(Out, encoding(utf8)),
+        write(Out, String),
+        set_stream(Out, encoding(octet))).
 http_post_data(cgi_stream(In, _Len), Out, HdrExtra) :-
     !,
     debug(obsolete, 'Obsolete 2nd argument in cgi_stream(In,Len)', []),
@@ -1075,7 +1139,7 @@ http_post_data(cgi_stream(In), Out, HdrExtra) :-
     content_length_in_encoding(Encoding, In, Size),
     http_join_headers(HdrExtra, Header, Hdr2),
     phrase(post_header(cgi_data(Size), Hdr2), HeaderText),
-    format(Out, '~s', [HeaderText]),
+    send_request_header(Out, HeaderText),
     setup_call_cleanup(
         set_stream(Out, encoding(Encoding)),
         copy_stream_data(In, Out),
@@ -1088,7 +1152,7 @@ http_post_data(form(Fields), Out, HdrExtra) :-
                       [ content_type('application/x-www-form-urlencoded')
                       ], Header),
     phrase(post_header(cgi_data(Size), Header), HeaderChars),
-    format(Out, '~s', [HeaderChars]),
+    send_request_header(Out, HeaderChars),
     format(Out, '~s', [Codes]).
 http_post_data(form_data(Data), Out, HdrExtra) :-
     !,
@@ -1106,7 +1170,7 @@ http_post_data(form_data(Data), Out, HdrExtra) :-
                               content_type(ContentType)
                             ], Header),
           phrase(post_header(cgi_data(Size), Header), HeaderChars),
-          format(Out, '~s', [HeaderChars]),
+          send_request_header(Out, HeaderChars),
           setup_call_cleanup(
               open_memory_file(MemFile, read, In, [encoding(octet)]),
               copy_stream_data(In, Out),
@@ -1130,7 +1194,7 @@ http_post_data(List, Out, HdrExtra) :-          % multipart-mixed
                               content_type(ContentType)
                             ], Header),
           phrase(post_header(cgi_data(Size), Header), HeaderChars),
-          format(Out, '~s', [HeaderChars]),
+          send_request_header(Out, HeaderChars),
           setup_call_cleanup(
               open_memory_file(MemFile, read, In, [encoding(octet)]),
               copy_stream_data(In, Out),
@@ -1176,6 +1240,11 @@ post_header(atom(Type, Atom), HdrExtra) -->
     header_fields(HdrExtra, Len),
     content_length(atom(Atom, utf8), Len),
     content_type(Type, utf8),
+    "\r\n".
+post_header(string(Type, String), HdrExtra) -->
+    header_fields(HdrExtra, Len),
+    content_length(string(String, utf8), Len),
+    content_type(Type, utf8),
     "\r\n".
 
 
@@ -1191,7 +1260,7 @@ post_header(atom(Type, Atom), HdrExtra) -->
 http_reply_header(Out, What, HdrExtra) :-
     phrase(reply_header(What, HdrExtra, _Code), String),
     !,
-    format(Out, '~s', [String]).
+    send_reply_header(Out, String).
 
 %!  reply_header(+Data, +HdrExtra, -Code)// is det.
 %
@@ -1316,6 +1385,7 @@ status_reply_headers(moved(To, Body), Body,
                      [ location(To) ]).
 status_reply_headers(moved_temporary(To, Body), Body,
                      [ location(To) ]).
+status_reply_headers(gone(_URL, Body), Body, []).
 status_reply_headers(see_other(To, Body), Body,
                      [ location(To) ]).
 status_reply_headers(authorise(Method, Body), Body,
@@ -1578,6 +1648,8 @@ content_length(Reply, Len) -->
 length_of(_, Len) :-
     nonvar(Len),
     !.
+length_of(string(String, Encoding), Len) :-
+    length_of(codes(String, Encoding), Len).
 length_of(codes(String, Encoding), Len) :-
     !,
     setup_call_cleanup(
@@ -1703,6 +1775,18 @@ read_field_value([H|T]) -->
     [H],
     read_field_value(T).
 
+%!  send_reply_header(+Out, +String) is det.
+%!  send_request_header(+Out, +String) is det.
+%
+%   Low level routines to send a single HTTP request or reply line.
+
+send_reply_header(Out, String) :-
+    debug(http(send_reply), "< ~s", [String]),
+    format(Out, '~s', [String]).
+
+send_request_header(Out, String) :-
+    debug(http(send_request), "> ~s", [String]),
+    format(Out, '~s', [String]).
 
 %!  http_parse_header_value(+Field, +Value, -Prolog) is semidet.
 %
