@@ -51,6 +51,9 @@
 #if defined(__linux__) && !defined(PATH_MAX)
 #include <linux/limits.h>
 #endif
+#ifndef PATH_MAX
+#define PATH_MAX 1024
+#endif
 
 static status	kindFile(FileObj f, Name encoding);
 static Sheet	FileFilters;
@@ -444,7 +447,8 @@ out:
 static status
 backup_name(const char *old, const char *ext, char *bak, size_t len)
 { if ( strlen(old) + strlen(ext) + 1 < len )
-  { sprintf(bak, "%s%s", old, ext);
+  { strcpy(bak, old);
+    strcat(bak, ext);
     succeed;
   } else
   { errno = ENAMETOOLONG;
@@ -552,7 +556,17 @@ getFilterFile(FileObj f)
       fail;
     }
 
-    sprintf(path, "%s%s", strName(f->name), strName(extension));
+    char *fname = strName(f->name);
+    char *ext   = strName(extension);
+
+    if ( strlen(fname)+strlen(ext)+1 >= sizeof(path) )
+    { errorPce(f, NAME_representation, NAME_nameTooLong);
+      fail;
+    }
+
+    strcpy(path, fname);
+    strcat(path, ext);
+
 #if O_XOS
     if ( _xos_exists(path, _XOS_FILE) )
 #else
@@ -911,7 +925,9 @@ getSizeFile(FileObj f)
 
 static Date
 getTimeFile(FileObj f, Name which)
-{
+{ if ( isDefault(which) )
+    which = NAME_modified;
+
 #if O_XOS
   Name name = getOsNameFile(f);
   double t;
@@ -922,9 +938,6 @@ getTimeFile(FileObj f, Name which)
 #else
   STAT_TYPE buf;
 
-  if ( isDefault(which) )
-    which = NAME_modified;
-
   if ( statFile(f, &buf) == 0 )
   { if ( which == NAME_modified )
       answer(CtoDate(buf.st_mtime));
