@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2022, University of Amsterdam
+    Copyright (c)  2011-2023, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -39,6 +39,7 @@
 #include "windows/uxnt.h"
 #include "config/wincfg.h"
 #include <winsock2.h>
+#include "../pl-nt.h"
 #define CRLF_MAPPING 1
 #else
 #include <config.h>
@@ -622,6 +623,7 @@ S__fillbuf(IOSTREAM *s)
   { char chr;
     ssize_t n;
 
+  again:
     n = (*s->functions->read)(s->handle, &chr, 1);
     if ( n == 1 )
     { c = char_to_int(chr);
@@ -631,7 +633,15 @@ S__fillbuf(IOSTREAM *s)
 	s->flags |= SIO_FEOF;
       return -1;
     } else
-    { S__seterror(s);
+    { if ( errno == EINTR )
+      { if ( PL_handle_signals() < 0 )
+	{ Sset_exception(s, PL_exception(0));
+	  errno = EPLEXCEPTION;
+	  return -1;
+	}
+	goto again;
+      }
+      S__seterror(s);
       return -1;
     }
   } else
@@ -658,6 +668,7 @@ S__fillbuf(IOSTREAM *s)
       len = s->bufsize;
     }
 
+  again2:
     n = (*s->functions->read)(s->handle, s->limitp, len);
     if ( n > 0 )
     { s->limitp += n;
@@ -668,6 +679,8 @@ S__fillbuf(IOSTREAM *s)
       { if ( !(s->flags & SIO_NOFEOF) )
 	  s->flags |= SIO_FEOF;
 	return -1;
+      } else if ( Sferror(s) )
+      { return -1;
 #ifdef EWOULDBLOCK
       } else if ( errno == EWOULDBLOCK )
       { s->bufp = s->buffer;
@@ -675,6 +688,13 @@ S__fillbuf(IOSTREAM *s)
 	S__seterror(s);
 	return -1;
 #endif
+      } else if ( errno == EINTR )
+      { if ( PL_handle_signals() < 0 )
+	{ Sset_exception(s, PL_exception(0));
+	  errno = EPLEXCEPTION;
+	  return -1;
+	}
+	goto again2;
       } else
       {	S__seterror(s);
 	return -1;
@@ -2172,6 +2192,21 @@ Sfprintf(IOSTREAM *s, const char *fm, ...)
   return rval;
 }
 
+/* SfprintfX() is identical to Sfprintf() but its definition in
+   SWI-Stream.h doesn't have the "check format" attribute. */
+
+int
+SfprintfX(IOSTREAM *s, const char *fm, ...)
+{ va_list args;
+  int rval;
+
+  va_start(args, fm);
+  rval = Svfprintf(s, fm, args);
+  va_end(args);
+
+  return rval;
+}
+
 
 int
 Sprintf(const char *fm, ...)
@@ -2269,7 +2304,8 @@ typedef enum
 { INT_INT       = 0,
   INT_LONG      = 1,
   INT_LONG_LONG = 2,
-  INT_SIZE_T    = 3
+  INT_SIZE_T    = 3,
+  INT_PTRDIFF_T = 4
 } int_type;
 
 int
@@ -2353,6 +2389,10 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    islong = INT_SIZE_T;
 	    fm++;
 	    break;
+	  case 't':
+	    islong = INT_PTRDIFF_T;
+	    fm++;
+	    break;
 	  case 'L':			/* %Ls: ISO-Latin-1 string */
 	    enc = ENC_ISO_LATIN_1;
 	    fm++;
@@ -2410,6 +2450,9 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	      case INT_SIZE_T:
 		vll = va_arg(args, size_t);
 		break;
+	      case INT_PTRDIFF_T:
+		vll = va_arg(args, ptrdiff_t);
+		break;
 	      default:
 		assert(0);
 	    }
@@ -2431,6 +2474,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 		break;
 	      case INT_LONG_LONG:
 	      case INT_SIZE_T:
+	      case INT_PTRDIFF_T:
 #ifdef __WINDOWS__
 		*fp++ = 'I';		/* Synchronise with INT64_FORMAT! */
 		*fp++ = '6';
@@ -2594,6 +2638,22 @@ Ssnprintf(char *buf, size_t size, const char *fm, ...)
 }
 
 
+/* SsnprintfX() is identical to Ssnprintf() but its definition in
+   SWI-Stream.h doesn't have the "check format" attribute. */
+
+int
+SsnprintfX(char *buf, size_t size, const char *fm, ...)
+{ va_list args;
+  int rval;
+
+  va_start(args, fm);
+  rval = Svsnprintf(buf, size, fm, args);
+  va_end(args);
+
+  return rval;
+}
+
+
 int
 Svsprintf(char *buf, const char *fm, va_list args)
 { IOSTREAM s;
@@ -2666,6 +2726,21 @@ Sdprintf(const char *fm, ...)
   return rval;
 }
 
+/* SdprintfX() is identical to Sdprintf() but its definition in
+   SWI-Stream.h doesn't have the "check format" attribute. */
+
+int
+SdprintfX(const char *fm, ...)
+{ va_list args;
+  int rval;
+
+  va_start(args, fm);
+  rval = Svdprintf(fm, args);
+  va_end(args);
+
+  return rval;
+}
+
 #ifdef O_DEBUG
 
 int
@@ -3484,6 +3559,32 @@ Sopen_iri_or_file(const char *path, const char *how)
   return s;
 }
 
+#ifdef __WINDOWS__
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+[not sure] It seems the CRT library of Windows can operate in two modes:
+one where the CRT state is  shared   between  the application and shared
+objects loaded into it, and one where this  is not the case. The default
+Windows build uses the shared view  and   the  Conda build the separated
+view.  This means that CRT handles cannot cross a component boundary.
+
+The clib/process.c component needs to turn an   I/O HANDLE into a Prolog
+stream.  As  Prolog  streams  use   the    CRT   handle,  we  must  call
+_open_osfhandle in the SWI-Prolog core.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+Swin_open_osfhandle(HANDLE h, int flags)
+{ return _open_osfhandle((intptr_t)h, flags);
+}
+
+IOSTREAM *
+Swin_open_handle(HANDLE h, const char *mode)
+{ int fd = Swin_open_osfhandle(h, 0);
+
+  return Sfdopen(fd, mode);
+}
+#endif
+
 
 IOSTREAM *
 Sfdopen(int fd, const char *type)
@@ -3549,6 +3650,17 @@ Sfileno(IOSTREAM *s)
 
 
 #ifdef __WINDOWS__
+
+HANDLE
+Swinhandle(IOSTREAM *s)
+{ int fd = Sfileno(s);
+
+  if ( fd >= 0 )
+    return (HANDLE)_get_osfhandle(fd);
+
+  return NULL;
+}
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 On  Windows,  type  SOCKET  is   an    unsigned   int   and  all  values
 [0..INVALID_SOCKET) are valid. It is  also   not  allowed  to run normal
@@ -3577,6 +3689,10 @@ Swinsock(IOSTREAM *s)
 		 *******************************/
 
 #ifdef HAVE_POPEN
+#ifdef __WINDOWS__
+#define popen _popen
+#define pclose _pclose
+#endif
 
 static ssize_t
 Sread_pipe(void *handle, char *buf, size_t size)
@@ -4110,12 +4226,12 @@ Sclose_buffer(IOSTREAM *s)
 		 *	 STANDARD HANDLES	*
 		 *******************************/
 
-#define STDIO(n, f) { NULL, NULL, NULL, NULL, \
-		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
-		      (void *)(n), &Sttyfunctions, \
-		      -1, \
-		      ENC_ISO_LATIN_1 \
-		    }
+#define STDIO(n, f) \
+	{ .bufp = NULL, .limitp = NULL, .buffer = NULL, .unbuffer = NULL, \
+	  .lastc = EOF, .magic = SIO_MAGIC, .flags = (f),		  \
+	  .handle = (void *)(n), .functions = &Sttyfunctions,		  \
+	  .timeout = -1, .encoding = ENC_ISO_LATIN_1			  \
+	}
 
 #define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY|SIO_TEXT)
 #define STDIO_STREAMS \
